/*

Copyright (C) 2002 - 2007 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <regex.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include <libnd.h>
#include <plugins/libnd_traceset.h>
#include <plugins/libnd_conntrack.h>
#include <protocols/tcp/libnd_tcp.h>


/* Interval in seconds between garbage collection of connection table */
#define LND_COUNT_GC_INTERVAL 3*60

typedef struct libnd_count {

  gboolean do_tcp;
  gboolean need_tcp_content;
  gboolean need_tcp_handshake;
  gboolean need_tcp_teardown;

  gboolean do_udp;
  gboolean need_udp_tx;
  gboolean need_udp_rx;
  int      udp_timeout;

  int      count;

  LND_Packet *current;
  LND_ConnTable *conns;

} LND_Count;

LND_Count *
count_new(void)
{
  LND_Count *count;

  if (! (count = g_new0(LND_Count, 1)))
    return NULL;

  if (! (count->conns = libnd_conn_table_new(LND_CONN_TABLE_IGNORE_DEAD)))
    {
      g_free(count);
      return NULL;
    }

  return count;
}

void
count_free(LND_Count *count)
{
  if (! count)
    return;

  libnd_conn_table_free(count->conns);
  g_free(count);
}

static void
count_help(void)
{
  printf("This plugin counts flows, letting you define what to consider a flow.\n"
	 "\n"
	 "USAGE: lndtool -r count [--tcp|-t] [--tcp-content|-T] [--udp|-u] [--udp-both|-U]\n"
	 "                        [--udp-timeout <NUM>] [--names-file|-f <FILE>] [<TRACE2> ...]\n"
	 "\n"
	 "  --help, -h, -?           This message.\n"
	 "  --tcp|t                  Count TCP flows.\n"
	 "  --tcp-content|T          Count only TCP flows that exchange content.\n"
	 "  --tcp-handshake          Count only TCP flows whose handshake is observed.\n"
	 "  --tcp-teardown           Count only TCP flows whose teardown is observed,\n"
	 "                           (Teardown is normal FINs or RST).\n"
	 "  --udp|u                  Count UDP flows.\n"
	 "  --udp-tx                 Count only UDP flows that see traffic going out.\n"
	 "  --udp-rx                 Count only UDP flows that see traffic coming back.\n"
	 "  --udp-timeout <NUM>      Use NUM seconds as the timeout for UDP flows.\n"
	 "  --names-file|-f <FILE>   Take filenames of traces from flatfile FILE.\n");
}


/* Garbage-collect the connection table.
 */
static gboolean
count_conn_table_gc_cb(LND_ConnID *conn, LND_Count *count)
{
  if (libnd_conn_is_dead(conn, count->current))
    {
      /* There is a possibilty of more than one connection in 
       * the table matching this connection -- to prevent segfaults
       * make sure we only nuke the one we mean.
       */
      if (libnd_conn_table_remove(count->conns, conn) == conn)
	libnd_conn_free(conn);
      
      return TRUE; /* Keep going. */
    }
  
  /* If the connection isn't dead yet, we abort the callback iteration,
   * because all future flows will be younger than the current one:
   * we're iterating backward through the age list.
   */
  return FALSE;
}

static struct bpf_timeval foo;

/* The main routine for a single trace -- iterates over all packets
 * in the given trace and maintains flow counts.
 */
static gboolean
count_trace(LND_TraceSet *set, LND_Trace *trace, LND_Count *count)
{
  LND_PacketIterator  pit;
  LND_Packet *packet;
  LND_ConnID *conn;
  struct bpf_timeval tv, last_gc;

  pcapnav_timeval_init(&last_gc, 0, 0);

  /* Iterate over entire individual trace, read-only. */
  for (libnd_pit_init_mode(&pit, trace, LND_PACKET_IT_AREA_R);
       libnd_pit_get(&pit); libnd_pit_next(&pit))
    {
      packet = libnd_pit_get(&pit);

      /* Make sure we have connection state for this packet's flow */
      if (! (conn = libnd_conn_table_lookup(count->conns, packet)))
	{
	  if (! (conn = libnd_conn_new(packet)))
	    /* It's a non-IP packet -- we ignore these. */
	    continue;
	  
	  libnd_conn_table_add(count->conns, conn);
	}
      
      /* We now definitely have connection state with associated flow data */
      libnd_conn_update(conn, packet);

      /* If the connection is dead, remove it right away and clean up
       * in order to keep the connection table as small as possible.
       */
      if (libnd_conn_is_dead(conn, packet))
	{
	  libnd_conn_table_remove(count->conns, conn);
	  libnd_conn_free(conn);

	  tv = packet->ph.ts;
	  continue;
	}

      if (! libnd_conn_data_get(conn, "counted"))
	{
	  gboolean do_count = FALSE;

	  if (conn->proto == IPPROTO_TCP && count->do_tcp)
	    {
	      LND_TCPConn *tcpc = (LND_TCPConn *) conn;

	      do_count = TRUE;

	      if (count->need_tcp_content && conn->content_tx == 0)
		do_count = FALSE;

	      if (count->need_tcp_handshake && ! libnd_tcpconn_handshake_seen(tcpc))
		do_count = FALSE;

	      if (count->need_tcp_teardown && ! libnd_tcpconn_teardown_seen(tcpc))
		do_count = FALSE;	      
	    }
	  
	  if (conn->proto == IPPROTO_UDP && count->do_udp)
	    {
	      do_count = TRUE;

	      if (count->need_udp_tx && conn->content_tx == 0)
		do_count = FALSE;	      

	      if (count->need_udp_rx && conn->content_rx == 0)
		do_count = FALSE;	      
	    }


	  if (do_count)
	    {
	      count->count++;
	      libnd_conn_data_set(conn, "counted", (void*) 1);
	    }
	}
            
      /* If we haven't garbage-collected the connection table in
       * LND_APPDEMUX_CONN_TABLE_GC_INTERVAL seconds, do so now.
       */
      pcapnav_timeval_sub(&packet->ph.ts, &last_gc, &tv);
      if (tv.tv_sec >= LND_COUNT_GC_INTERVAL)
	{
	  libnd_conn_table_foreach_oldest(count->conns, (LND_ConnFunc) count_conn_table_gc_cb, count);
	  last_gc = packet->ph.ts;
	}
    }
  
  return TRUE;
  TOUCH(set);
}

/* ---- Plugin symbols below ------------------------------------------- */

const char *
name(void)
{
  return "Count";
}

const char *
author(void)
{
  return "Christian Kreibich, <christian@whoop.org>";
}

const char *
version(void)
{
  return "0.1";
}


gboolean
init(void)
{
  if (! libnd_plugin_find("Trace-Set"))
    {
      D(("Trace-Set plugin not found\n"));
      return FALSE;
    }

  if (! libnd_plugin_find("Conntrack"))
    {
      D(("Conntrack plugin not found\n"));
      return FALSE;
    }

  D(("Count plugin initialization complete.\n"));
  
  return TRUE;
}

gboolean
run(LND_Trace *trace, LND_PluginArgs *args)
{
  LND_Count *count;
  LND_TraceSet *traces;
  int i, files = 0;

  if (! (traces = libnd_traceset_new()))
    return FALSE;

  if (! (count = count_new()))
    {
      libnd_traceset_free(traces);
      return FALSE;
    }

  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  count_help();
	  return FALSE;
	}
      else if (!strcmp(args->argv[i], "-f") ||
	       !strcmp(args->argv[i], "--names-file"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the name of a file that contains\n"
		     "a list of trace files, one filename per line, when\n"
		     "using the --names-file|-f option.\n");
	      return FALSE;
	    }
	  
	  if (! libnd_misc_can_read(args->argv[i]))
	    printf("Could not open '%s'\n", args->argv[i]);
	  else if (! libnd_traceset_add_trace_name_list(traces, args->argv[i]))
	    printf("Could not add all traces provided in '%s'\n", args->argv[i]);
	  else
	    files++;
	}
      else if (!strcmp(args->argv[i], "-t") || !strcmp(args->argv[i], "--tcp"))
	{
	  count->do_tcp = TRUE;
	}
      else if (!strcmp(args->argv[i], "-T") || !strcmp(args->argv[i], "--tcp-content"))
	{
	  count->need_tcp_content = TRUE;
	}
      else if (!strcmp(args->argv[i], "--tcp-handshake"))
	{
	  count->need_tcp_handshake = TRUE;
	}
      else if (!strcmp(args->argv[i], "--tcp-teardown"))
	{
	  count->need_tcp_teardown = TRUE;
	}
      else if (!strcmp(args->argv[i], "-u") || !strcmp(args->argv[i], "--udp"))
	{
	  count->do_udp = TRUE;
	}
      else if (!strcmp(args->argv[i], "--udp-tx"))
	{
	  count->need_udp_tx = TRUE;
	}
      else if (!strcmp(args->argv[i], "--udp-rx"))
	{
	  count->need_udp_rx = TRUE;
	}
      else if (!strcmp(args->argv[i], "--udp-timeout"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the UDP timeout you want to use to\n"
		     "the --udp-timeout option, in seconds.\n");
	      return FALSE;
	    }
	 
	  count->udp_timeout = atoi(args->argv[i]);
	  libnd_conntrack_generic_timeout = count->udp_timeout;
	}
      else if (args->argv[i][0] != '-')
	{
	  if (! libnd_traceset_add_trace_name(traces, args->argv[i]))
	    printf("Trace file %s not added.\n", args->argv[i]);
	  else
	    files++;
	}
      else
	printf("Warning: option %s unknown.\n", args->argv[i]);
    }
  
  if (args->argc == 0)
    {
      count_help();
      return FALSE;
    }

  libnd_traceset_foreach(traces, (LND_TraceSetCB) count_trace, count);
  libnd_traceset_free(traces);

  printf("%i\n", count->count);

  count_free(count);

  return TRUE;
  TOUCH(trace);
}
