/*

Copyright (C) 2002 - 2007 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <libnd.h>

const char *
name(void) { return "IAT"; }

const char *
author(void) { return "Christian Kreibich <christian@whoop.org>"; }

const char *
version(void) { return "0.1"; }

gboolean
init(void) { return TRUE; }

static void
iat_help(void)
{
  printf("This plugin reports packet inter-arrival times (IATs). It can use a BPF\n"
	 "filtering expression to focus on subsets of packets, report accumulative\n"
	 "or original timestamps of packets along with the IATs, and can also include\n"
	 "tcpdump output alongside.\n"
	 "\n"
	 "USAGE: lndtool -r iat [--help|-h|-?] [--filter|-f <filter>] [--tcpdump|-t]\n"
	 "                      [--orig-times|-o] [--acc-times|-a] <trace file>\n"
	 "\n"
	 "  --help, -h, -?           This message.\n"
	 "  --filter|-f <filter>     A BPF filter expression, a la tcpdump/pcap.\n"
	 "  --tcpdump|-t             Show tcpdump output alongside the output.\n"
	 "  --orig-times|-o          Report original timestamps of packets.\n"
	 "  --acc-times|-a           Report accumulative timestamps of packets.\n"
	 "\n"
	 "If --tcpdump|-t is used, any additional options are used to modify\n"
	 "tcpdump's output; otherwise, they are ignored.\n");
}


gboolean
run(LND_Trace *unused, LND_PluginArgs *args)
{
  LND_Trace *trace = NULL;
  LND_PacketIterator  pit;
  LND_Packet *packet;
  pcap_t *pcap;
  int i;
  const char *filter = NULL;
  struct bpf_timeval tv_last;
  struct bpf_timeval tv_acc = { 0, 0 };
  struct bpf_program prog;
  struct bpf_insn *fcode = NULL;
  gboolean do_tcpdump = FALSE;
  gboolean do_acc = FALSE;
  gboolean do_orig = FALSE;

  memset(&prog, 0, sizeof(struct bpf_program));

  /* Make sure that tcpdump, if it ends up being used, doesn't display
   * a timestamp.
   */
  libnd_tcpdump_print_timestamp = FALSE;

  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  iat_help();
	  return FALSE;
	}
      else if (!strcmp(args->argv[i], "-f") ||
	       !strcmp(args->argv[i], "--filter"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass a pcap filter expression when\n"
		     "using the --f|--filter option.\n");
	      return FALSE;
	    }
	  
	  filter = args->argv[i];
	}
      else if (!strcmp(args->argv[i], "-a") ||
	       !strcmp(args->argv[i], "--acc-times"))
	do_acc = TRUE;
      else if (!strcmp(args->argv[i], "-o") ||
	       !strcmp(args->argv[i], "--orig-times"))
	do_orig = TRUE;
      else if (!strcmp(args->argv[i], "-t") ||
	       !strcmp(args->argv[i], "--tcpdump"))
	do_tcpdump = TRUE;
      else
	{
	  if (! trace)
	    trace = libnd_trace_new(args->argv[i]);
	}
    }
  
  if (args->argc == 0 || ! trace)
    {
      iat_help();
      return FALSE;
    }

  pcap = libnd_trace_get_pcap_handle(trace);
  
  if (filter && pcap_compile(pcap, &prog, (char*) filter, 0, 0) < 0)
    {
      printf("Error compiling filter expression: %s\n", pcap_geterr(pcap));
      return FALSE;
    }

  /* Initialize timestamps using first packet. */
  libnd_pit_init_mode(&pit, trace, LND_PACKET_IT_AREA_R);
  packet = libnd_pit_get(&pit);
  libnd_pit_next(&pit);

  tv_last.tv_sec = packet->ph.ts.tv_sec;
  tv_last.tv_usec = packet->ph.ts.tv_usec;
  
  for ( ; libnd_pit_get(&pit); libnd_pit_next(&pit))
    {
      if ( (fcode = prog.bf_insns) == NULL ||
	   bpf_filter(fcode, packet->data, packet->ph.len, packet->ph.caplen))
	{
	  struct bpf_timeval delta;

	  /* Compute delta to last packet's timestamp */
	  pcapnav_timeval_sub(&packet->ph.ts, &tv_last, &delta);

	  /* Add delta to accumulative time count */
	  pcapnav_timeval_add(&tv_acc, &delta, &tv_acc);

	  /* Print this IAT */
	  printf("%lu.%06lu ",
		 (long unsigned) delta.tv_sec,
		 (long unsigned) delta.tv_usec);

	  if (do_acc)
	    printf("%lu.%06lu ",
		   (long unsigned) tv_acc.tv_sec,
		   (long unsigned) tv_acc.tv_usec);
	  
	  if (do_orig)
	    printf("%lu.%06lu ",
		   (long unsigned) packet->ph.ts.tv_sec,
		   (long unsigned) packet->ph.ts.tv_usec);
	  
	  if (do_tcpdump)
	    {
	      char buf[1024];
	      libnd_tcpdump_get_packet_line(packet, buf, 1024, FALSE);
	      printf("%s", buf);
	    }
	    
	  printf("\n");

	  tv_last.tv_sec = packet->ph.ts.tv_sec;
	  tv_last.tv_usec = packet->ph.ts.tv_usec;
	}
    }
  
  return TRUE;
  TOUCH(unused);
}
