/*

Copyright (C) 2002 - 2007 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/
#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <sys/resource.h>
#include <sys/wait.h>

#include <unistd.h>
#include <limits.h>
#include <dirent.h>
#include <errno.h>

#include <libnd.h>
#include <libnd_misc.h>
#include <plugins/libnd_mux.h>

#define LIBND_MUX_BATCHSIZE 10

LND_Mux *
libnd_mux_new(void)
{
  LND_Mux *mux;

  if (! (mux = g_new0(LND_Mux, 1)))
    return NULL;
  
  if (! (mux->traces = libnd_traceset_new()))
    {
      D(("out of memory"));
      g_free(mux);
      return NULL;
    }

  mux->mode = LND_MUX_MODE_MERGE;
  mux->output_file = strdup("-");
  
  return mux;
}


void
libnd_mux_free(LND_Mux *mux)
{
  if (! mux)
    return;
  
  fflush(stdout);
  libnd_traceset_free(mux->traces);

  g_free(mux->output_file);
  
  if (mux->output_pcap)
    pcap_close(mux->output_pcap);

  if (mux->output_dumper)
    pcap_dump_close(mux->output_dumper);
  
  g_free(mux);
}


gboolean
libnd_mux_set_output(LND_Mux *mux, const char *output_file)
{
  if (! mux || !output_file || !*output_file)
    return FALSE;
    
  g_free(mux->output_file);
  mux->output_file = g_strdup(output_file);
  
  return TRUE;
}


static gboolean
mux_caplen_cb(LND_TraceSet *set, const char *name, LND_Mux *mux)
{
  int caplen;
  pcap_t *pcap = pcap_open_offline(name, libnd_pcap_errbuf);

  if (! pcap)
    return TRUE;

  caplen = pcap_snapshot(pcap);
  
  if (caplen > mux->max_caplen)
    {
      mux->max_caplen = caplen;
      
      if (mux->output_pcap)
	pcap_close(mux->output_pcap);

      mux->output_pcap = pcap;
    }
  else
    pcap_close(pcap);

  if (! mux->search_max_caplen)
    return FALSE;
  
  return TRUE;
  TOUCH(set);
}

static gboolean
mux_determine_caplen(LND_Mux *mux)
{
  if (! mux)
    return FALSE;
  
  /* We need to do a scan across all the files to see what is
   * the largest capture size we need. Since we may be writing to
   * stdout, we cannot just update it as we go and at the end seek
   * back and correct it.
   */

  /* Reset maximum caplen before scanning... */
  mux->max_caplen = 0;

  /* Iterate over all input traces and output dem packets. */
  libnd_traceset_foreach_name(mux->traces, (LND_TraceSetNameCB) mux_caplen_cb, mux);

  if (mux->output_pcap)
    mux->output_dumper = pcap_dump_open(mux->output_pcap, mux->output_file);

  return (mux->max_caplen != 0 && mux->output_dumper);
}


typedef struct lnd_mux_traceset_data {

  pcap_t **pcap_handles;
  int pcap_handles_index;
  int num_slots_alive;

} LND_MuxTracesetData;

static gboolean
mux_open_cb(LND_TraceSet *set, const char *name, LND_MuxTracesetData *data)
{
  pcap_t *pcap = pcap_open_offline(name, libnd_pcap_errbuf);

  if (! pcap)
    {
      printf("Couldn't open %s\n", name);
      data->num_slots_alive--;
      return TRUE;
    }

  data->pcap_handles[data->pcap_handles_index++] = pcap;
  
  return TRUE;
  TOUCH(set);
}

static gboolean
mux_delete_temps_cb(LND_TraceSet *set, const char *name, void *unused)
{
  unlink(name);

  return TRUE;
  TOUCH(unused);
  TOUCH(set);
}

static void
mux_process_batch(LND_Mux *mux, LND_TraceSet *set, const char* file)
{
  int i,best_slot;
  LND_Packet *packet;
  struct bpf_timeval tbest;
  LND_MuxTracesetData data;
  pcap_dumper_t *dumper;
  LND_Packet **packets;
  int num_slots;

  num_slots = libnd_traceset_get_size(set);
  packets = calloc(num_slots, sizeof(LND_Packet *));

  memset(&data, 0, sizeof(LND_MuxTracesetData));
  data.num_slots_alive = num_slots;
  data.pcap_handles = calloc(num_slots, sizeof(pcap_t *));
  data.pcap_handles_index = 0;

  libnd_traceset_foreach_name(set, (LND_TraceSetNameCB) mux_open_cb, &data);
    
  if (! (dumper = pcap_dump_open(mux->output_pcap, file)))
    {
      printf("Couldn't dump to %s\n", file);
      goto cleanup_return;
    }

  while (data.num_slots_alive > 0)
    {
      /* Read one round of packets (one from each child), build a list of those
       * packets in chronologically ordered sequence, and write out that
       * list.
       */
      pcapnav_timeval_init(&tbest, 0, 0);
      best_slot = -1;

      for (i = 0; i < num_slots; i++)
	{
	  if (! data.pcap_handles[i])
	    continue;
	  
	  if (! packets[i])
	    packets[i] = libnd_packet_from_pcap(data.pcap_handles[i]);
	  
	  if (! packets[i])
	    {
	      data.num_slots_alive--;
	      pcap_close(data.pcap_handles[i]);
	      data.pcap_handles[i] = NULL;
	      continue;
	    }
	  
	  if (tbest.tv_sec == 0 || pcapnav_timeval_cmp(&(packets[i]->ph.ts), &tbest) < 0)
	    {
	      tbest = packets[i]->ph.ts;
	      best_slot = i;
	    }
	}
      
      if (best_slot >= 0)
	{
	  packet = packets[best_slot];

	  libnd_packet_dump(packet, dumper);
	  libnd_packet_free(packet);
	  packets[best_slot] = NULL;
	}
    }

 cleanup_return:

  if (dumper)
    pcap_dump_close(dumper);

  for (i = 0; i < num_slots; i++)
    {
      if (data.pcap_handles[i])
	pcap_close(data.pcap_handles[i]);

      if (packets[i])
	libnd_packet_free(packets[i]);
    }
  
  g_free(packets);
  g_free(data.pcap_handles);
}

static void
mux_chron(LND_Mux *mux)
{
  GList *l, *partitions;
  int iteration = 0;
  int batch = 0;
  char buf[MAXPATHLEN];
  LND_TraceSet *current_set, *next_set;

  if (! mux)
    return;

  if (! mux_determine_caplen(mux))
    return;

  current_set = mux->traces;
  next_set = NULL;

  do
    {
      int num_partitions = libnd_traceset_get_size(current_set) / LIBND_MUX_BATCHSIZE + 1;

      if (! (partitions = libnd_traceset_partition(current_set, num_partitions)))
	{
	  D(("Cannot partition set of traces.\n"));
	  goto cleanup_return;
	}
            
      if (! (next_set = libnd_traceset_new()))
	goto cleanup_return;

      for (batch = 0, l = partitions; l; l = g_list_next(l), batch++)
	{
	  LND_TraceSet *set = l->data;

	  g_snprintf(buf, MAXPATHLEN, "__lnd_mux_tmp.%i.%i", iteration, batch);
	  libnd_traceset_add_trace_name(next_set, buf);
	  mux_process_batch(mux, set, buf);

	  if (current_set != mux->traces)
	    libnd_traceset_foreach_name(set, (LND_TraceSetNameCB) mux_delete_temps_cb, NULL);
	}
      
      for (l = partitions; l; l = g_list_next(l))
	libnd_traceset_free(l->data);
      g_list_free(partitions);

      if (current_set != mux->traces)
	libnd_traceset_free(current_set);
      
      current_set = next_set;
      next_set = NULL;

      iteration++;

    } while (libnd_traceset_get_size(current_set) > 1);


  if (rename(buf, mux->output_file) < 0)
    {
      printf("Could not rename %s to %s.\n", buf, mux->output_file);
    }

 cleanup_return:

  if (current_set != mux->traces)
    libnd_traceset_free(current_set);
}

static gboolean
mux_timeshift_cb(LND_TraceSet *set, LND_Trace *trace, LND_Mux *mux)
{
  LND_PacketIterator  pit;
  LND_Packet *packet;
  struct bpf_timeval diff, delta;
  int sign = 1;

  if (mux->verbose)
    {
      mux->trace_num++;
      fprintf(stderr, "-- Processing %ith file\n", mux->trace_num);
    }

  libnd_pit_init_mode(&pit, trace, LND_PACKET_IT_AREA_R);
  if (! (packet = libnd_pit_get(&pit)))
    {
      /* Seems to be an empty trace, not much to do then! */
      return TRUE;
    }
  
  if (mux->tv.tv_sec == 0 && mux->tv.tv_usec == 0)
    {
      pcapnav_timeval_init(&diff, 0, 0);
      sign = 1;
    }
  else
    {
      /* If the user provided a delta, add that. Otherwise just
       * increase by 1 microsecond so the next packet really does
       * come right after the last one.
       */
      if (mux->delta_secs)
	pcapnav_timeval_init(&delta, mux->delta_secs, 0);
      else
	pcapnav_timeval_init(&delta, 0, 1);
      

      /* We want to start from mux->tv, but we're at packet->ph.ts,
       * so figure out the difference and adjust each packet's
       * timestamp accordingly.
       */
      if (pcapnav_timeval_cmp(&mux->tv, &packet->ph.ts) < 0)
	{
	  pcapnav_timeval_sub(&packet->ph.ts, &mux->tv, &diff);
	  pcapnav_timeval_sub(&diff, &delta, &diff);
	  sign = -1;
	}
      else
	{
	  pcapnav_timeval_sub(&mux->tv, &packet->ph.ts, &diff);
	  pcapnav_timeval_add(&diff, &delta, &diff);
	  sign = +1;
	}
      
      /* diff is now absolute difference */
    }
  
  /* Iterate over entire individual trace, read-only. */
  for ( ; libnd_pit_get(&pit); libnd_pit_next(&pit))
    {
      packet = libnd_pit_get(&pit);
      
      /* Adjust the packet's timestamp... */
      if (sign > 0)
	pcapnav_timeval_add(&packet->ph.ts, &diff, &packet->ph.ts);
      else
	pcapnav_timeval_sub(&packet->ph.ts, &diff, &packet->ph.ts);
      
      /* Check the new timestamp against the last one. If there
       * are intra-trace non-chronological jumps, we can iron them
       * out now.
       */
      if (pcapnav_timeval_cmp(&packet->ph.ts, &mux->tv) <= 0)
	{
	  /* Bring it up to the same level... */
	  packet->ph.ts.tv_sec = mux->tv.tv_sec;
	  packet->ph.ts.tv_usec = mux->tv.tv_usec;

	  /* ... and add user-specified delta, if given,
	   * or the minimum.
	   */
	  if (mux->delta_secs)
	    packet->ph.ts.tv_sec += mux->delta_secs;
	  else
	    packet->ph.ts.tv_usec++;
	}
      
      /* Write it out. */
      libnd_packet_dump(packet, mux->output_dumper);
      
      /* Keep track of progress of time. */
      mux->tv = packet->ph.ts;
    }

  return TRUE;
  TOUCH(set);
}

static void
mux_timeshift(LND_Mux *mux)
{
  if (! mux_determine_caplen(mux))
    {
      fprintf(stderr, "Could not determine maximum capture length.\n");
      return;
    }

  pcapnav_timeval_init(&mux->tv, 0, 0);
  
  /* Iterate over all input traces and output dem packets. */
  libnd_traceset_foreach(mux->traces, (LND_TraceSetCB) mux_timeshift_cb, mux);
}


static gboolean
mux_seq_cb(LND_TraceSet *set, pcap_t *pcap, LND_Mux *mux)
{
  if (mux->verbose)
    {
      mux->trace_num++;
      fprintf(stderr, "-- Processing %ith file\n", mux->trace_num);
    }

  /* Iterate over entire individual trace */
  pcap_loop(pcap, -1, pcap_dump, (u_char *) mux->output_dumper);
  
  return TRUE;
  TOUCH(set);
}

static void
mux_seq(LND_Mux *mux)
{
  if (! mux_determine_caplen(mux))
    {
      fprintf(stderr, "Could not determine maximum capture length.\n");
      return;
    }

  /* Iterate over all input traces and output dem packets. */
  libnd_traceset_foreach_pcap(mux->traces, (LND_TraceSetPcapCB) mux_seq_cb, mux);
}


void
libnd_mux(LND_Mux *mux)
{
  if (!mux)
    return;

  switch (mux->mode)
    {
    case LND_MUX_MODE_MERGE:
      mux_chron(mux);
      break;

    case LND_MUX_MODE_TIMESHIFT:
      mux_timeshift(mux);
      break;

    case LND_MUX_MODE_CONCAT:
    default:
      mux_seq(mux);
    }
  
  /* Clean up dumper */
  pcap_dump_close(mux->output_dumper);
  mux->output_dumper = NULL;
}


const char *
name(void)
{
  return "Mux";
}


const char *
author(void)
{
  return "Christian Kreibich <christian.kreibich-AT-cl.cam.ac.uk>";
}


const char *
version(void)
{
  return VERSION;
}


gboolean
init(void)
{
  if (! libnd_plugin_find("Trace-Set"))
    {
      fprintf(stderr, "Trace-Set plugin not found\n");
      return FALSE;
    }

  return TRUE;
}


static void
mux_help(void)
{
  printf("Flow Multiplexer plugin\n"
	 "USAGE: lndtool -r mux [--help|-h] [--debug|-d] [--as-given|-a] [--merge|-m]\n"
	 "               [--timeshift|-t [--delta <secs]] [--output|-o FILE]\n"
	 "               [--names-file|-f FILE] <TRACE1> [<TRACE2> <TRACE3> ...]\n"
	 "\n"
	 "  --help, -h, -?           This message.\n"
	 "  --debug, -d              Debugging output.\n"
	 "  --verbose, -v            Be verbose\n"
	 "\n"
	 "  --as-given|-a            Just concatenate traces in sequence provided.\n"
	 "  --merge|-m               Fully multiplex input traces, establishing the\n"
	 "                           sequence of packets given by their timestamps.\n"
	 "                           This is the default.\n"
	 "  --timeshift|-t           Concatenate input traces, but adjust start time\n"
	 "                           of each trace so it follows the one before it.\n"
	 "                           This mode will also patch up any non-chronological\n"
	 "                           intra-trace jumps in time.\n"
	 "  --delta <secs>           For use with --timeshift|-t: number of seconds\n"
	 "                           by which to separate a trace. Default: 1usec,\n"
	 "                           i.e., the traces are concatenated with the smallest\n"
	 "                           temporal distance possible.\n"
	 "\n"
	 "  --output|-o FILE         Write output to FILE. Default is stdout.\n"
	 "  --names-file|-f FILE     Take filenames of traces from flatfile FILE.\n");
}

int
run(LND_Trace *trace, LND_PluginArgs *args)
{
  int i, counter = 0;
  LND_Mux *mux;
  
  if (!args)
    return 1;
  
  if (args->argc == 0)
    {
      printf("Please provide one or more traces to scan.\n");
      mux_help();
      return 1;
    }
  
  mux = libnd_mux_new();
  
  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  mux_help();
	  return 1;
	}
      else if (!strcmp(args->argv[i], "-o") ||
	       !strcmp(args->argv[i], "--output"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to provide a filename to the --output|-o option.\n");
	      return 1;
	    }

	  if (! libnd_mux_set_output(mux, args->argv[i]))
	    {
	      printf("Cannot use %s as output file: %s.\n", args->argv[i], strerror(errno));
	      return 1;
	    }
	}
      else if (!strcmp(args->argv[i], "-f") ||
	       !strcmp(args->argv[i], "--names-file"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the name of a file that contains\n"
		     "a list of trace files, one filename per line, when\n"
		     "using the --names-file|-f option.\n");
	      return 1;
	    }
	  
	  if (! libnd_traceset_add_trace_name_list(mux->traces, args->argv[i]))
	    printf("Could not add all traces provided in '%s'\n", args->argv[i]);
	  else
	    counter++;
	}
      else if (!strcmp(args->argv[i], "--delta"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the number of seconds by which to\n"
		     "separate consecutive traces to --delta, in seconds.\n");
	      return 1;
	    }

	  mux->delta_secs = atoi(args->argv[i]);
	}
      else if (!strcmp(args->argv[i], "-a") ||
	       !strcmp(args->argv[i], "--as-given"))
	{
	  mux->mode = LND_MUX_MODE_CONCAT;
	}
      else if (!strcmp(args->argv[i], "-t") ||
	       !strcmp(args->argv[i], "--timeshift"))
	{
	  mux->mode = LND_MUX_MODE_TIMESHIFT;
	}
      else if (!strcmp(args->argv[i], "-m") ||
	       !strcmp(args->argv[i], "--merge"))
	{
	  mux->mode = LND_MUX_MODE_MERGE;
	}
      else if (!strcmp(args->argv[i], "-v") ||
	       !strcmp(args->argv[i], "--verbose"))
	{
	  mux->verbose = TRUE;
	}
      else if (args->argv[i][0] != '-')
	{
	  if (! libnd_traceset_add_trace_name(mux->traces, args->argv[i]))
	    printf("Trace file %s not added.\n", args->argv[i]);
	  else
	    counter++;
	}
    }

  if (counter > 0)
    libnd_mux(mux);

  libnd_mux_free(mux);
  
  return 0;
  TOUCH(trace);
}
